/*

* Copyright (c) 2006, The Hong Kong University of Science and Technology (HKUST)
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without 
* modification, are permitted provided that the following conditions 
* are met:
*
*   * Redistributions of source code must retain the above copyright notice, 
*     this list of conditions,the authors and the following disclaimer.
*   * Redistributions in binary form must reproduce the above copyright notice,
*     this list of conditions,the authors and the following disclaimer in
*     the documentation and/or other materials provided with the distribution.
*   * Neither the name of the university nor the names of its 
*     contributors may be used to endorse or promote products derived from 
*     this software without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
* AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
* IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
* ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
* LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
* CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
* SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
* INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
* ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
* POSSIBILITY OF SUCH DAMAGE.
*
*/

import java.io.*;
import java.util.*;
import java.net.*;

/**
 * This class implements the backbone of our Educt simulator, an 
 * Event-orienteD qUery proCessing framework for sensor neTworks.
 *                                                   
 * @author Wenwei Xue (wwxue@cs.ust.hk)
 * @version 0.11
 */
 
public class Educt {
	
	static final byte GRID = 0x01;
	static final byte STAR = 0x02;
	static final byte INTELLAB = 0x03;
	byte topologyType = STAR;	
	/** Number of rows and columns in the network toplogy. */
	static int m = 1, n = 2; // m*n is the total number of nodes in the simulated sensor network (including Node 0)	
	static final int hopDistance = 10;  // unit: meter	
	
	Node eductNetwork [][] = new Node [m][n];
	int terminatedNodeNum = 0;  // indicating how many nodes in the simulated network has terminated 
		
	static String virtualTableName = "sensors";
	/** Default size of an attribute */
	static final int ATTR_SIZE = 2;  // unit: byte
	static String attributes [] = {"nodeid", "temp", "light", "humidity", "noise", "accel_x", "accel_y", "mag_x", "mag_y", "pressure", "voltage", "loc_x", "loc_y", "depth"};	
	/* Aggregate queries are not supported in the current version of Educt. */
	//static String aggregates [] = {"max", "min", "count", "sum", "avg"};		
		
	/** Names of the system built-in methods for pattern matching */
	static final String PATTERN_METHOD_NAME = "pattern";
	static final String HORIZON_METHOD_NAME = "horizon";
	static final String SLOPE_METHOD_NAME = "slope";
	static final String OSCILLATION_METHOD_NAME = "oscillation";		
	static final String JUMP_METHOD_NAME = "jump";
	static final String SPIKE_METHOD_NAME = "spike";	
	
	// read the query workload from this file
	static String queryFile = "../conf/system/queries.txt";
	static final int numQueries = 1;  // # concurrent queries run in the simulation
	 
	// the directory where the configuration files of the methods are put
	static String confFilePathPrefix = "../conf/users/";
		
	// Points to a system-level log file that records some overall information about the simulation
	PrintWriter overallLogFileWriter;	
	
	Educt ()
	  {
	  	try
	      {
	      	String logFileName = Query.logPathPrefix + "all.log";
	      	overallLogFileWriter = new PrintWriter(new FileWriter(logFileName));
	      }
	    catch (Exception e) {e.printStackTrace();}    
	  	
	  	System.out.println("Simulated sensor network topology (" + m + "*" + n + " grid):");
	  	System.out.println();
	  		  
	    overallLogFileWriter.println("Simulated sensor network topology (" + m + "*" + n + " grid):");
	  	overallLogFileWriter.println();
	  	overallLogFileWriter.flush();
	  	
	  	System.out.println("Network topology (nodeid):");
	  	
	  	overallLogFileWriter.println("Network topology (nodeid):");
	  	overallLogFileWriter.flush();
	  	
	  	int id = 0;
	  	int i,j;
	  	for (i=0;i<m;i++)
	  	  {
	  	  	for (j=0;j<n;j++)
	  	      {
	  	        eductNetwork[i][j] = new Node(this, id++);	  	        
	  	        
	  	        eductNetwork[i][j].loc_x = i*hopDistance;
	  	        eductNetwork[i][j].loc_y = j*hopDistance;
	  	        
	  	        eductNetwork[i][j].depth = -1;	  	        
	  	        
	  	        if (id <= 10)
	  	          {System.out.print(" "); overallLogFileWriter.print(" ");}
	  	        
	  	        System.out.print(eductNetwork[i][j].id);
	  	        // System.out.print("(" + eductNetwork[i][j].loc_x + "," + eductNetwork[i][j].loc_y + ")");
	  	        System.out.print(" ");
	  	        
	  	        overallLogFileWriter.print(eductNetwork[i][j].id);
	  	        overallLogFileWriter.print(" ");
	  	      }
	  	     
	  	     System.out.println();
	  	     
	  	     overallLogFileWriter.println();
	  	     overallLogFileWriter.flush();	  	     
	  	  } 
	  	  
	    System.out.println();
	    System.out.println("Network topology (depth):");
	    
	    overallLogFileWriter.println();
	    overallLogFileWriter.println("Network topology (depth):");
	    eductNetwork[0][0].depth = 0;
	    
	    System.out.print(" 0" + " ");
	    
	    overallLogFileWriter.print(" 0" + " ");
	    overallLogFileWriter.flush();
	    	    
	    for (i=0;i<m;i++)
	      {
	      	for (j=0;j<n;j++)
	          {
	            if (eductNetwork[i][j].depth != -1)
	              continue;
	              
	            int depth = Integer.MAX_VALUE;
	              
	            if (topologyType == STAR)
	              {
	              	depth = eductNetwork[i][j].depth = 1;
	          
	                System.out.print(" ");
	                System.out.print(depth + " ");
	                
	                overallLogFileWriter.print(" ");
	                overallLogFileWriter.print(depth + " ");  
	                
	                continue;
	              }  
	              
	            if (topologyType == GRID) // assume a node can communicate with at most 8 neighbors nearby
	              {
	              	if (j-1 >= 0)
	              	  if (eductNetwork[i][j-1].depth != -1)
	                    if (eductNetwork[i][j-1].depth + 1 < depth)
	                  	  depth = eductNetwork[i][j-1].depth+1;
	                
	            	if (i-1 >= 0)
	              	  if (eductNetwork[i-1][j].depth != -1)
	                	if (eductNetwork[i-1][j].depth + 1 < depth)
	                  	  depth = eductNetwork[i-1][j].depth+1;
	                
	            	if (j+1 < n)
	              	  if (eductNetwork[i][j+1].depth != -1)
	                	if (eductNetwork[i][j+1].depth + 1 < depth)
	                  	  depth = eductNetwork[i][j+1].depth+1; 
	                     
	            	if (i+1 < m)
	              	  if (eductNetwork[i+1][j].depth != -1)
	                	if (eductNetwork[i+1][j].depth + 1 < depth)
	                  	  depth = eductNetwork[i+1][j].depth+1;	
	                
	            	if ((i-1 >= 0) && (j-1 >= 0))
	              	  if (eductNetwork[i-1][j-1].depth != -1)
	                	if (eductNetwork[i-1][j-1].depth + 1 < depth)
	                  	  depth = eductNetwork[i-1][j-1].depth+1; 
	                
	            	if ((i-1 >= 0) && (j+1 < n))
	              	  if (eductNetwork[i-1][j+1].depth != -1)
	                	if (eductNetwork[i-1][j+1].depth + 1 < depth)
	                  	  depth = eductNetwork[i-1][j+1].depth+1; 
	          
	            	if ((i+1 < m) && (j-1 >= 0))
	              	  if (eductNetwork[i+1][j-1].depth != -1)
	                	if (eductNetwork[i+1][j-1].depth + 1 < depth)
	                  	  depth = eductNetwork[i+1][j-1].depth+1; 
	                
	            	if ((i+1 < m) && (j+1 < n))
	             	  if (eductNetwork[i+1][j+1].depth != -1)
	                	if (eductNetwork[i+1][j+1].depth + 1 < depth)
	                  	  depth = eductNetwork[i+1][j+1].depth+1;
	                
	            	eductNetwork[i][j].depth = depth;
	          
	            	System.out.print(" ");
	           	 	System.out.print(depth + " ");   
	            	
	            	overallLogFileWriter.print(" ");
	            	overallLogFileWriter.print(depth + " "); 
	            	
	            	continue;
	         	  }
	         	  
	         	if (topologyType==INTELLAB)
	         	  {
	         	   	if (eductNetwork[i][j].id <= 8)
	         	  	  eductNetwork[i][j].depth = 1;
	         	  	else
	         	  	  if (eductNetwork[i][j].id <= 11)
	         	  	    eductNetwork[i][j].depth = 2;
	         	  	  else
	         	  	    if (eductNetwork[i][j].id <= 14)
	         	  	      eductNetwork[i][j].depth = 3;
	         	  	    else
	         	  	      if (eductNetwork[i][j].id == 15 || eductNetwork[i][j].id == 16 || eductNetwork[i][j].id == 18)
	         	  	        eductNetwork[i][j].depth = 4;
	         	  	      else
	         	  	        if (eductNetwork[i][j].id == 19)
	         	  	          eductNetwork[i][j].depth = 6;
	         	  	        else if (eductNetwork[i][j].id == 17 || eductNetwork[i][j].id <= 22)
	         	  	          	   eductNetwork[i][j].depth = 5;
	         	  	        else
	         	  	          if (eductNetwork[i][j].id <= 38)
	         	  	            eductNetwork[i][j].depth = 1;
	         	  	          else if (eductNetwork[i][j].id <= 41)
	         	  	                 eductNetwork[i][j].depth = 2;
	         	  	               else
	         	  	                 if (eductNetwork[i][j].id <= 43)
	         	  	                   eductNetwork[i][j].depth = 3;
	         	  	                 else
	         	  	                   if (eductNetwork[i][j].id == 44)
	         	  	                     eductNetwork[i][j].depth = 4;
	         	  	                   else
	         	  	                     if (eductNetwork[i][j].id == 45)
	         	  	                       eductNetwork[i][j].depth = 1;
	         	  	                     else
	         	  	                       if (eductNetwork[i][j].id <= 47 || eductNetwork[i][j].id == 49 || eductNetwork[i][j].id == 51)
	         	  	                         eductNetwork[i][j].depth = 2;
	         	  	                       else
	         	  	                         if (eductNetwork[i][j].id == 48 || eductNetwork[i][j].id == 52)
	         	  	                           eductNetwork[i][j].depth = 3;
	         	  	                         else
	         	  	                           if (eductNetwork[i][j].id == 50 || eductNetwork[i][j].id == 53)
	         	  	                             eductNetwork[i][j].depth = 4;
	         	  	                           else
	         	  	                             eductNetwork[i][j].depth = 5;	         	  	                             	         	  	      
	         	  	  	
	         	  	depth = eductNetwork[i][j].depth;
	         	  	
	         	  	System.out.print(" ");
	           	 	System.out.print(depth + " ");   
	            	
	            	overallLogFileWriter.print(" ");
	            	overallLogFileWriter.print(depth + " "); 	         	  	
	         	  }	            	             	
	          }
	          
	        System.out.println();
	        
	        overallLogFileWriter.println();
	      }	
	      
	    System.out.println();    
	    
	    overallLogFileWriter.println();	    
	  	
	  }  //Educt
	
	
	public boolean checkValidAttribute (String name)
	  {
	  	int i;
	  	for (i=0;i<attributes.length;i++)
	  	  if (attributes[i].equals(name))
	  	    return true;
	  	    
	  	return false;
	  		  	
	  }  //checkValidAttribute
	
	
	public int convertUnit(String str)
	  {
	  	if (str.equals("second"))
	  	  return 1;
	  	
	  	if (str.equals("minute"))
	  	  return 60;
	  	  
	  	if (str.equals("hour"))
	  	  return 60*60;
	  	  
	  	if (str.equals("day"))
	  	  return 60*60*24;	  	
	  	
	  	else return 1;
	  		  	
	  }  //convertTimeUnit
	  
	String parseTag (String configString, String tag)
	  {
	  	int start = configString.indexOf("<"+tag+">");
	  	if (start == -1)
	  	  {
	  	  	 System.err.println("Invalid method configuration file format. Tag <"+tag+"> is missing.");
	  	  	return null;
	  	  }
	  	start += ("<"+tag+">").length();
	  	int stop = configString.indexOf("</"+tag+">",start);
	  	if (stop == -1)
	  	  {
	  	  	System.err.println("Invalid method configuration file format. Tag </"+tag+"> is missing.");
	  	  	return null;
	  	  }
	   	return configString.substring(start,stop).trim();
	  
	  }  //parseTag
	  
	double parseValueTag (String configString, String tag)
	  {
	  	int start = configString.indexOf("<"+tag+">");
	  	if (start == -1)
	  	 {
	  	   System.err.println("Invalid method configuration file format. Tag <intervalofInspection> is missing.");
	  	   	return -1;
	  	 }      
	  	start += ("<"+tag+">").length();
	  	int stop = configString.indexOf(" ",start);
	  	if (stop == -1)
	  	  stop = configString.indexOf("\t",start);
	  	if (stop == -1)
	  	  {
	  	  	System.err.println("Invalid method configuration file format. Fail to parse tag <intervalofInspection></intervalofInspection>.");
	  	   	return -1;
	  	  }	  	  	
	  	 double value = Double.parseDouble(configString.substring(start,stop).trim());
	  	 start = stop + 1;
	  	 stop = configString.indexOf("</"+tag+">",start);
	  	 if (stop == -1)
	  	  {
	  	  	System.err.println("Invalid method configuration file format. Fail to parse tag <intervalofInspection></intervalofInspection>.");
	  	  	return -1;
	  	  }
	  	if (value > 1) stop--;  	
	  	value *= convertUnit(configString.substring(start,stop).trim());
	  	
	  	return value;
	  	
	  }  //parseValueTag    
    
    public Query parseQuery(int queryID,String queryString)
	  {
	  	Query q = new Query(queryID, queryString);
	  	
	  	int start, stop, end, mark;
	  	String str;
	  	
	  	// SELECT clause
        start = queryString.indexOf("select ");
        if (start == -1)
          {
            System.err.println("Invalid query syntax: SELECT clause is missing.");
            return null;
          }
        start += "select ".length();
        
        end = queryString.indexOf("from ",start);
        if (end == -1)
          {
            System.err.println("Invalid query syntax: FROM clause is missing.");
            return null;
          }	
          
        while (start<end)
          {
          	stop = queryString.indexOf(",",start);
          	if (stop>end || stop==-1)
          	  stop = end;
          	
          	if (!checkValidAttribute(str=queryString.substring(start,stop).trim()))
          	  {
          	  	System.err.println("Invalid attribute: " + str + " in the SELECT clause.");
          	  	return null;
          	  }
          	else
          	  q.attributes[q.attrNum++] = str;          	
          	start = stop + 1; 
          }  	
	  	
	  	/* for (int i=0;i<q.attrNum;i++)
	  	     System.out.println(q.attributes[i]);
	  	   System.exit(0);
	  	 */	  	
	  	
	  	// FROM clause
	  	start = end + "from ".length();
	  	end = queryString.indexOf("where ", start);
	  	if (end == -1)
          {
            /** We require each query in Educt must have a WHERE clause */
            System.err.println("Invalid query syntax: WHERE clause is missing.");
            return null;
          }
        
        str = queryString.substring(start,end).trim();
        if (!virtualTableName.equals(str))
          {
          	System.err.println("Invalid virtual table name " + str + " in the FROM clause: must be \"sensors\".");
            return null;
          }
        start = end + "where ".length();
        mark = start;        
        
        // WHERE clause	
	  	/** Currently we assume only one method embedded in the WHERE clause */
	  	start = queryString.indexOf(PATTERN_METHOD_NAME,mark);
	  	if (start == -1)
	  	  start = queryString.indexOf(HORIZON_METHOD_NAME,mark);
	  	if (start == -1)
	  	  start = queryString.indexOf(SLOPE_METHOD_NAME,mark);
	  	if (start == -1)
	  	  start = queryString.indexOf(OSCILLATION_METHOD_NAME,mark);
	  	if (start == -1)
	  	  start = queryString.indexOf(JUMP_METHOD_NAME,mark);
	  	if (start == -1)
	  	  start = queryString.indexOf(SPIKE_METHOD_NAME,mark);
	  	if (start == -1)
	  	  {
	  	  	start = queryString.indexOf("where ");
	  	  	start += "where ".length();
	  	    start -= "and ".length();
	  	    while (start != -1)
	  	      {
	  	  	    start += "and ".length();
	  	  	    while (queryString.charAt(start) == ' ')
	  	  	      start ++;
	  	  	    stop = queryString.indexOf(" ",start);
	  	  	    if (stop == -1)
	  	  	      {
	  	  	  	    System.err.println("Invalid syntax in the WHERE clause.");
	  	  	  	    return null;
	  	  	      }
	  	  	    str = queryString.substring(start, stop).trim();
	  	  	    if (!checkValidAttribute(str))
	  	  	      {
	  	  	  	    System.err.println("Invalid attribute " + str + " in the WHERE clause.");
	  	  	  	    return null;
	  	  	      } 	  	    
	  	        Predicate pred = new Predicate();
	  	        pred.attribute = str;	  	    
	  	        start += str.length();
	  	        while (queryString.charAt(start) == ' ')
	  	  	      start ++;
	  	    	stop = queryString.indexOf(" ",start);	 
	  	  	    if (stop == -1)
	  	  	      {
	  	  	  	    System.err.println("Invalid syntax in the WHERE clause.");
	  	  	  	    return null;
	  	  	      }
	  	  	    str = queryString.substring(start, stop).trim();
	  	  	    if (!str.equals(">") && !str.equals("<") && !str.equals(">=") && 
	  	  	        !str.equals("<=") && !str.equals("=") && !str.equals("!="))
	  	  	      {
	  	  	      	System.err.println("Invalid operator " + str + " in the WHERE clause.");
	  	  	     	return null;	  	  	  	
	  	  	      }  
	  	  	    pred.op = str;
	  	  	    start += str.length();
	  	  	    while (queryString.charAt(start) == ' ')
	  	  	      start ++;
	  	  	    stop = queryString.indexOf(" ",start);	 
	  	  	    if (stop == -1)
	  	  	      {
	  	  	  	    System.err.println("Invalid syntax in the WHERE clause.");
	  	  	  	    return null;
	  	  	      }
	  	     	str = queryString.substring(start, stop).trim();
	  	        try {double d = Double.parseDouble(str);}
	  	        catch (Exception e)
	  	  	      {
	  	  	  	    System.err.println("Invalid constant value " + str + " in the WHERE clause.");
	  	  	  	    return null;	  	  	  	
	  	  	      }  
	  	  	    pred.constant = Double.parseDouble(str);
	  	  	    q.predicates.addElement((Predicate)pred);	  	  	    
	  	  	    start = queryString.indexOf("and ",start);
	  	      }
	  	
	  	    q.type = Query.THRESHOLD;
	  	    
	  	    q.M = null;  // threshold-based query does not have an embedded method
	  	    
	  	    return q;	  	      	  	
	  	  }
	  	
	  	q.type = Query.EVENT_DRIVEN;
	  	stop = queryString.indexOf("(",start);
	  	EDMethod M = new EDMethod();
	  	M.name = queryString.substring(start,stop).trim();	  	
	  	M.id = queryID;
	  	start = stop + 1;
	  	stop = queryString.indexOf(",",start);
	  	if (stop == -1)
	  	  {
	  	  	System.err.println("Invalid method syntax.");
	  	  	return null;
	  	  }
	  	M.attribute = queryString.substring(start,stop).trim();
	  	start = stop + 1;
	  	stop = queryString.indexOf(")",start);
	  	if (stop == -1)
	  	  {
	  	  	System.err.println("Invalid method syntax.");
	  	  	return null;
	  	  }
	  	M.configFile = confFilePathPrefix + queryString.substring(start+2,stop-1).trim();
	  	// System.out.println(q.M.id + "\t" + q.M.name + "\t" + q.M.attribute + "\t" + q.M.configFile);
	  		  	
	  	String configString = "";
	  	BufferedReader configFileReader;
	  	try
	  	  {
	  	    configFileReader = new BufferedReader(new FileReader(M.configFile));
	  	    while ((str = configFileReader.readLine()) != null)
	  	      {
	  	  	    configString += str;
	  	  	    configString += "\r\n";
	  	      }
	  	    configFileReader.close();
	  	    // System.out.println(configString);
	  	  }
	  	catch (Exception e) {e.printStackTrace();}
	  	
	  	
	  	if (M.name.equals(PATTERN_METHOD_NAME))
	  	  {
	  	  	PatternMethod pattern = new PatternMethod(q);
	  	  	pattern.name = M.name;
	  	  	pattern.id = M.id;
	  	  	pattern.attribute = M.attribute;
	  	  	pattern.configFile = M.configFile;
	  	  	
	  	  	q.M = pattern;
	  	  	
	  	  	start = configString.indexOf("<samplePeriod>");
	  	  	if (start == -1)
	  	  	  pattern.sp = Query.DEFAULT_SAMPLE_PERIOD;
	  	  	else
	  	  	  {
	  	  	  	pattern.sp = (int)parseValueTag(configString,"samplePeriod");
	  	  		if (pattern.sp < 0)
	  	  	      {
	  	  	  		System.err.println("Invalid method configuration file format. Fail to parse tag <samplePeriod></samplePeriod>.");
	  	  	  		return null;
	  	  	  	  }	  	  		
	  	      }
	  	  	  
	  	   	pattern.T = (int)parseValueTag(configString,"intervalofInspection");
	  	  	if (pattern.T < 0)
	  	  	  {
	  	  	  	System.err.println("Invalid method configuration file format. Fail to parse tag <intervalofInspection></intervalofInspection>.");
	  	  	  	return null;
	  	  	  }	  	  	
	  	  	pattern.n = pattern.T / pattern.sp + 1;
	  	  	pattern.span = pattern.T + 2*pattern.sp;  // the two sample periods is set for buffering purpose
	  	  		  	
	  	  	pattern.delta = parseValueTag(configString,"degreeofDifference");
	  	  	
	  	  	try {pattern.one_minus_alpha = Double.parseDouble(parseTag(configString, "confidenceLevel"));}
	  	  	catch (Exception e) {e.printStackTrace();}	  	  	  
	  	  	if (pattern.one_minus_alpha <= 0 || pattern.one_minus_alpha >= 1)
	  	  	  {
	  	  	    System.err.println("Invalid method configuration file format. Fail to parse tag <confidenceLevel></confidenceLevel>.");
	  	  	    return null;
	  	  	  }
	  	  	
	  	  	pattern.epsilon = parseValueTag(configString, "errorBound");
	  	  		  	  	
	  	  	start = configString.indexOf("<querySequence>");
	  	  	if (start == -1)
	  	  	  {
	  	  	    System.err.println("Invalid method configuration file format. Tag <querySequence> is missing.");	  	  	    
	  	  	    return null;
	  	  	  }
	  	  	start += "<querySequence>".length();
	  	  	end = configString.indexOf("</querySequence>",start);
	  	  	if (end == -1)
	  	  	  {
	  	  	    System.err.println("Invalid method configuration file format. Tag </querySequence> is missing.");	  	  	    
	  	  	    return null;
	  	  	  }	  	  	
	  	  	start = configString.indexOf("<value>",start);
	  	  	if (start == -1)	  	  	  
	  	  	  {
	  	  	    System.err.println("Invalid method configuration file format. Fail to parse tag <querySequence></querySequence>.");	  	  	    
	  	  	    return null;
	  	  	  }	  	  		  	
	  	  	while (start!=-1&&start < end)
	  	  	  {
	  	  	  	start += "<value>".length();
	  	  	  	stop = configString.indexOf("</value>",start);
	  	  	  	if (stop == -1)
	  	  	  	  {
	  	  	    	System.err.println("Invalid method configuration file format. Fail to parse tag <querySequence></querySequence>.");	  	  	    
	  	  	    	return null;
	  	  	  	  }
	  	  	  	pattern.Q[pattern.len++] = (int)Double.parseDouble((configString.substring(start,stop)).trim());
	  	  	  	start = configString.indexOf("<value>",start);	  	  	  	
	  	  	  }  	  		  	  	
	  	  }
	  	  
	  	if (M.name.equals(HORIZON_METHOD_NAME))
	  	  {
	  	  	HorizonMethod horizon = new HorizonMethod(q);
	  	  	horizon.name = M.name;	  	  	
	  	  	horizon.id = M.id;
	  	  	horizon.attribute = M.attribute;
	  	  	horizon.configFile = M.configFile;
	  	  	
	  	  	q.M = horizon;
	  	  	
	  	  	if (!horizon.name.equals(parseTag(configString, "type")))
	  	  	  {
	  	  	  	System.err.println("Invalid method configuration file format. Fail to parse tag <type></type>.");
	  	  	  	return null;
	  	  	  }
	  	  	
	  	  	start = configString.indexOf("<samplePeriod>");
	  	  	if (start == -1)
	  	  	  horizon.sp = Query.DEFAULT_SAMPLE_PERIOD;
	  	  	else
	  	  	  {
	  	  	  	horizon.sp = (int)parseValueTag(configString,"samplePeriod");
	  	  		if (horizon.sp < 0)
	  	  	      {
	  	  	  		System.err.println("Invalid method configuration file format. Fail to parse tag <samplePeriod></samplePeriod>.");
	  	  	  		return null;
	  	  	  	  }	  	  		
	  	      }
	  	  	  
	  	   	horizon.T = (int)parseValueTag(configString,"intervalofInspection");
	  	  	if (horizon.T < 0)
	  	  	  {
	  	  	  	System.err.println("Invalid method configuration file format. Fail to parse tag <intervalofInspection></intervalofInspection>.");
	  	  	  	return null;
	  	  	  }	  	  	
	  	  	horizon.n = horizon.T / horizon.sp + 1;
	  	  	horizon.span = horizon.T + 2*horizon.sp;  // the two sample periods is set for buffering purpose
	  	  		  	
	  	    try {horizon.one_minus_alpha = Double.parseDouble(parseTag(configString, "confidenceLevel"));}
	  	  	catch (Exception e) {e.printStackTrace();}		  	  	  
	  	  	if (horizon.one_minus_alpha <= 0 || horizon.one_minus_alpha >= 1)
	  	  	  {
	  	  	    System.err.println("Invalid method configuration file format. Fail to parse tag <confidenceLevel></confidenceLevel>.");
	  	  	    return null;
	  	  	  }
	  	  	
	  	  	horizon.epsilon = parseValueTag(configString, "errorBound");	  	  	 	  	
	  	  }	  	
	  	
	  	if (M.name.equals(SLOPE_METHOD_NAME))
	  	  {
	  	  	SlopeMethod slope = new SlopeMethod(q);
	  	  	slope.name = M.name;	  	  	
	  	  	slope.id = M.id;
	  	  	slope.attribute = M.attribute;
	  	  	slope.configFile = M.configFile;
	  	  	
	  	  	q.M = slope;	  	  	
	  	  	
	  	   	if (!slope.name.equals(parseTag(configString, "type")))
	  	  	  {
	  	  	  	System.err.println("Invalid method configuration file format. Fail to parse tag <type></type>.");
	  	  	  	return null;
	  	  	  }  	  	
	  	  	
	  	  	str = parseTag(configString, "mode");
	  	  	if (str.equals("increasing"))
	  	  	  slope.mode = EDMethod.I;
	  	  	else if (str.equals("decreasing"))
	  	  	       slope.mode = EDMethod.D;
	  	  	     else
	  	  	       {
	  	  	         System.err.println("Invalid method configuration file format. Fail to parse tag <mode></mode>.");
	  	  	         return null;
	  	  	       }	  	  	
	  	  	
	  	  	start = configString.indexOf("<samplePeriod>");
	  	  	if (start == -1)
	  	  	  slope.sp = Query.DEFAULT_SAMPLE_PERIOD;
	  	  	else
	  	  	  {
	  	  	  	slope.sp = (int)parseValueTag(configString,"samplePeriod");
	  	  		if (slope.sp < 0)
	  	  	      {
	  	  	  		System.err.println("Invalid method configuration file format. Fail to parse tag <samplePeriod></samplePeriod>.");
	  	  	  		return null;
	  	  	  	  }	  	  		
	  	      }
	  	  	  
	  	   	slope.T = (int)parseValueTag(configString,"intervalofInspection");
	  	  	if (slope.T < 0)
	  	  	  {
	  	  	  	System.err.println("Invalid method configuration file format. Fail to parse tag <intervalofInspection></intervalofInspection>.");
	  	  	  	return null;
	  	  	  }	  	  	
	  	  	slope.n = slope.T / slope.sp + 1;
	  	  	slope.span = slope.T + 2*slope.sp;  // the two sample periods is set for buffering purpose
	  	  	
	  	  	if (slope.mode == EDMethod.I)
	  	  	  try {slope.one_minus_alpha1 = Double.parseDouble(parseTag(configString, "nonDecreasingConfidenceLevel"));}
	  	  	  catch (Exception e) {e.printStackTrace();}	  	  	  
	  	  	else
	  	  	  	try {slope.one_minus_alpha1 = Double.parseDouble(parseTag(configString,"nonIncreasingConfidenceLevel"));}
	  	  	    catch (Exception e) {e.printStackTrace();}   	  	  
	  	  	if (slope.one_minus_alpha1 <= 0 || slope.one_minus_alpha1 >= 1)
	  	  	 {
	  	  	   System.err.println("Invalid method configuration file format. Fail to parse tag <nonIncreasingConfidenceLevel></nonIncreasingConfidenceLevel>.");
	  	  	   return null;
	  	  	 }	  	  	  	
	  	  	  
	  	  	if (slope.mode == EDMethod.I)
	  	  	  try {slope.one_minus_alpha2 = Double.parseDouble(parseTag(configString, "nonDecreasingConfidenceLevel"));}
	  	  	  catch (Exception e) {e.printStackTrace();}	  	  	  
	  	  	else
	  	  	  	try {slope.one_minus_alpha2 = Double.parseDouble(parseTag(configString,"nonIncreasingConfidenceLevel"));}
	  	  	    catch (Exception e) {e.printStackTrace();}
	  	  	if (slope.one_minus_alpha2 <= 0 || slope.one_minus_alpha2 >= 1)
	  	  	  {
	  	  	  	System.err.println("Invalid method configuration file format. Fail to parse tag <increasingConfidenceLevel></increasingConfidenceLevel>.");
	  	  	    return null;
	  	  	  } 	  
	  	  		
	  	  	slope.delta = parseValueTag(configString,"degreeofChange");	  	  	 	  	
	  	  	
	  	  	slope.epsilon = parseValueTag(configString,"errorBound"); 
	  	  }
	  	  
	  	if (M.name.equals(OSCILLATION_METHOD_NAME))
	  	  {
	  	  	OscillationMethod oscillation = new OscillationMethod(q);
	  	  	oscillation.name = M.name;	  	  	
	  	  	oscillation.id = M.id;
	  	  	oscillation.attribute = M.attribute;
	  	  	oscillation.configFile = M.configFile;
	  	  	
	  	  	q.M = oscillation;	  	  	
	  	  	
	  	   	if (!oscillation.name.equals(parseTag(configString, "type")))
	  	  	  {
	  	  	  	System.err.println("Invalid method configuration file format. Fail to parse tag <type></type>.");
	  	  	  	return null;
	  	  	  } 	  	  	
	  	  	
	  	  	start = configString.indexOf("<samplePeriod>");
	  	  	if (start == -1)
	  	  	  oscillation.sp = Query.DEFAULT_SAMPLE_PERIOD;
	  	  	else
	  	  	  {
	  	  	  	oscillation.sp = (int)parseValueTag(configString,"samplePeriod");
	  	  		if (oscillation.sp < 0)
	  	  	      {
	  	  	  		System.err.println("Invalid method configuration file format. Fail to parse tag <samplePeriod></samplePeriod>.");
	  	  	  		return null;
	  	  	  	  }	  	  		
	  	      }
	  	  	  
	  	   	oscillation.T = (int)parseValueTag(configString,"intervalofInspection");
	  	   	if (oscillation.T < 0)
	  	  	  {
	  	  	  	System.err.println("Invalid method configuration file format. Fail to parse tag <intervalofInspection></intervalofInspection>.");
	  	  	  	return null;
	  	  	  }	  	  	
	  	  	oscillation.n = oscillation.T / oscillation.sp + 1;
	  	  	oscillation.span = oscillation.T + 2*oscillation.sp;  // the two sample periods is set for buffering purpose
	  	  	
	  	   	try {oscillation.one_minus_alpha = Double.parseDouble(parseTag(configString, "confidenceLevel"));}
	  	  	catch (Exception e) {e.printStackTrace();}		  	  	  
	  	  	if (oscillation.one_minus_alpha <= 0 || oscillation.one_minus_alpha >= 1)
	  	  	  {
	  	  	    System.err.println("Invalid method configuration file format. Fail to parse tag <confidenceLevel></confidenceLevel>.");
	  	  	    return null;
	  	  	  }
	  	  	
	  	  	try {oscillation.lou = Double.parseDouble(parseTag(configString, "percentageofTransience"));}
	  	  	catch (Exception e) {e.printStackTrace();}		  	  	  
	  	  	if (oscillation.lou <= 0 || oscillation.lou >= 1)
	  	  	  {
	  	  	    System.err.println("Invalid method configuration file format. Fail to parse tag <percentageofTransience></percentageofTransience>.");
	  	  	    return null;
	  	  	  }	  	  	
	  	  	
	  	  	oscillation.epsilon = parseValueTag(configString, "errorBound");  		  	  	  	  	
	  	  }
	  	
	  	if (M.name.equals(JUMP_METHOD_NAME))
	  	  {
	  	  	JumpMethod jump = new JumpMethod(q);
	  	  	jump.name = M.name;	  	  	
	  	  	jump.id = M.id;
	  	  	jump.attribute = M.attribute;
	  	  	jump.configFile = M.configFile;
	  	  	
	  	  	q.M = jump;
	  	  	
	  	  	if (!jump.name.equals(parseTag(configString,"type")))
	  	  	  {
	  	  	  	System.err.println("Invalid method configuration file format. Fail to parse tag <type></type>.");
	  	  	  	return null;
	  	  	  }	  	  	
	  	  	
	  	  	str = parseTag(configString,"mode");
	  	  	if (str.equals("increasing"))
	  	  	  jump.mode = EDMethod.I;
	  	  	else if (str.equals("decreasing"))
	  	  	       jump.mode = EDMethod.D;
	  	  	     else
	  	  	       {
	  	  	         System.err.println("Invalid method configuration file format. Fail to parse tag <mode></mode>.");
	  	  	         return null;
	  	  	       }	  	  	
	  	  	
	  	  	start = configString.indexOf("<samplePeriod>");
	  	  	if (start == -1)
	  	  	  jump.sp = Query.DEFAULT_SAMPLE_PERIOD;
	  	  	else
	  	  	  {
	  	  	  	jump.sp = (int)parseValueTag(configString,"samplePeriod");
	  	  		if (jump.sp < 0)
	  	  	      {
	  	  	  		System.err.println("Invalid method configuration file format. Fail to parse tag <samplePeriod></samplePeriod>.");
	  	  	  		return null;
	  	  	  	  }	  	  		
	  	      }
	  	  	  
	  	  	jump.T1 = (int)parseValueTag(configString,"anteriorIntervalofInspection");	  	  	
	  	  	if (jump.T1 < 0)
	  	  	  {
	  	  	  	System.err.println("Invalid method configuration file format. Fail to parse tag <anteriorIntervalofInspection></anteriorIntervalofInspection>.");
	  	  	  	return null;
	  	  	  }	  	  	
	  	  	jump.n1 = jump.T1 / jump.sp + 1;
	  	  	
	  	  	jump.T2 = (int)parseValueTag(configString,"posteriorIntervalofInspection");	  	  	
	  	  	if (jump.T2 < 0)
	  	  	  {
	  	  	  	System.err.println("Invalid method configuration file format. Fail to parse tag <posteriorIntervalofInspection></posteriorIntervalofInspection>.");
	  	  	  	return null;
	  	  	  }	  	  	
	  	  	jump.n2 = jump.T2 / jump.sp + 1;
	  	  	jump.T = jump.T1 + jump.T2 + jump.sp;
	  	  	jump.n = jump.T / jump.sp + 1;
	  	  	jump.span = Math.max(jump.T1,jump.T2) + jump.sp + 2*jump.sp;	  	  	
	  	  	  
	  	  	jump.delta1 = parseValueTag(configString, "degreeofInstantaneousChange");
	  	  	jump.delta2 = parseValueTag(configString, "degreeofLevelChange");  	  	
	  	  	
	  	  	try {jump.one_minus_alpha = Double.parseDouble(parseTag(configString, "confidenceLevel"));}
	  	  	catch (Exception e) {e.printStackTrace();}			  	  	  
	  	  	if (jump.one_minus_alpha <= 0 || jump.one_minus_alpha >= 1)
	  	  	  {
	  	  	    System.err.println("Invalid method configuration file format. Fail to parse tag <confidenceLevel></confidenceLevel>.");
	  	  	    return null;
	  	  	  }
	  	  	
	  	  	jump.epsilon = parseValueTag(configString, "errorBound");  	  		  
	  	  }
	  	  
	  	if (M.name.equals(SPIKE_METHOD_NAME))
	  	  {
	  	  	SpikeMethod spike = new SpikeMethod(q);
	  	  	spike.name = M.name;	  	  	
	  	  	spike.id = M.id;
	  	  	spike.attribute = M.attribute;
	  	  	spike.configFile = M.configFile;
	  	  	
	  	  	q.M = spike;
	  	  	
	  	  	if (!spike.name.equals(parseTag(configString,"type")))
	  	  	  {
	  	  	  	System.err.println("Invalid method configuration file format. Fail to parse tag <type></type>.");
	  	  	  	return null;
	  	  	  }	  	  	
	  	  	
	  	  	str = parseTag(configString,"mode");
	  	  	if (str.equals("increasing"))
	  	  	  spike.mode = EDMethod.I;
	  	  	else if (str.equals("decreasing"))
	  	  	       spike.mode = EDMethod.D;
	  	  	     else
	  	  	       {
	  	  	         System.err.println("Invalid method configuration file format. Fail to parse tag <mode></mode>.");
	  	  	         return null;
	  	  	       }	  	  	
	  	  	
	  	  	start = configString.indexOf("<samplePeriod>");
	  	  	if (start == -1)
	  	  	  spike.sp = Query.DEFAULT_SAMPLE_PERIOD;
	  	  	else
	  	  	  {
	  	  	  	spike.sp = (int)parseValueTag(configString,"samplePeriod");
	  	  		if (spike.sp < 0)
	  	  	      {
	  	  	  		System.err.println("Invalid method configuration file format. Fail to parse tag <samplePeriod></samplePeriod>.");
	  	  	  		return null;
	  	  	  	  }	  	  		
	  	      }
	  	  	  
	  	  	spike.T1 = (int)parseValueTag(configString,"anteriorIntervalofInspection");	  	  	
	  	  	if (spike.T1 < 0)
	  	  	  {
	  	  	  	System.err.println("Invalid method configuration file format. Fail to parse tag <anteriorIntervalofInspection></anteriorIntervalofInspection>.");
	  	  	  	return null;
	  	  	  }	  	  	
	  	  	spike.n1 = spike.T1 / spike.sp + 1;
	  	  	
	  	  	spike.T2 = (int)parseValueTag(configString,"posteriorIntervalofInspection");	  	  	
	  	  	if (spike.T2 < 0)
	  	  	  {
	  	  	  	System.err.println("Invalid method configuration file format. Fail to parse tag <posteriorIntervalofInspection></posteriorIntervalofInspection>.");
	  	  	  	return null;
	  	  	  }	  	  	
	  	  	spike.n2 = spike.T2 / spike.sp + 1;
	  	  	spike.T = spike.T1 + spike.T2 + spike.sp;
	  	  	spike.n = spike.T / spike.sp + 1;
	  	  	spike.span = Math.max(spike.T1,spike.T2) + spike.sp + 2*spike.sp;	  	  	
	  	  	  
	  	  	spike.delta1 = parseValueTag(configString, "degreeofInstantaneousChange");
	  	  	spike.delta2 = parseValueTag(configString, "degreeofLevelChange");  	  	
	  	  	
	  	  	try {spike.one_minus_alpha = Double.parseDouble(parseTag(configString, "confidenceLevel"));}
	  	  	catch (Exception e) {e.printStackTrace();}		  	  	  
	  	  	if (spike.one_minus_alpha <= 0 || spike.one_minus_alpha >= 1)
	  	  	  {
	  	  	    System.err.println("Invalid method configuration file format. Fail to parse tag <confidenceLevel></confidenceLevel>.");
	  	  	    return null;
	  	  	  }
	  	  	
	  	  	spike.epsilon = parseValueTag(configString, "errorBound");  
	  	  }	  	  
	  	
	  	start = queryString.indexOf("where ");
	  	start = queryString.indexOf("and ",start);
	  	while (start != -1)
	  	  {
	  	  	start += "and ".length();
	  	  	while (queryString.charAt(start) == ' ')
	  	  	  start ++;
	  	  	stop = queryString.indexOf(" ",start);
	  	  	if (stop == -1)
	  	  	  {
	  	  	  	System.err.println("Invalid syntax in the WHERE clause.");
	  	  	  	return null;
	  	  	  }
	  	  	str = queryString.substring(start, stop).trim();
	  	  	if (!checkValidAttribute(str))
	  	  	  {
	  	  	  	System.err.println("Invalid attribute " + str + " in the WHERE clause.");
	  	  	  	return null;
	  	  	  } 	  	    
	  	    Predicate pred = new Predicate();
	  	    pred.attribute = str;	  	    
	  	    start += str.length();
	  	    while (queryString.charAt(start) == ' ')
	  	  	  start ++;
	  	  	stop = queryString.indexOf(" ",start);	 
	  	  	if (stop == -1)
	  	  	  {
	  	  	  	System.err.println("Invalid syntax in the WHERE clause.");
	  	  	  	return null;
	  	  	  }
	  	  	str = queryString.substring(start, stop).trim();
	  	  	if (!str.equals(">") && !str.equals("<") && !str.equals(">=") && 
	  	  	    !str.equals("<=") && !str.equals("=") && !str.equals("!="))
	  	  	  {
	  	  	  	System.err.println("Invalid operator " + str + " in the WHERE clause.");
	  	  	  	return null;	  	  	  	
	  	  	  }  
	  	  	pred.op = str;
	  	  	start += str.length();
	  	  	while (queryString.charAt(start) == ' ')
	  	  	  start ++;
	  	  	stop = queryString.indexOf(" ",start);	 
	  	  	if (stop == -1)
	  	  	  {
	  	  	  	System.err.println("Invalid syntax in the WHERE clause.");
	  	  	  	return null;
	  	  	  }
	  	  	str = queryString.substring(start, stop).trim();
	  	    try {double d = Double.parseDouble(str);}
	  	    catch (Exception e)
	  	  	  {
	  	  	  	System.err.println("Invalid constant value " + str + " in the WHERE clause.");
	  	  	  	return null;	  	  	  	
	  	  	  }  
	  	  	pred.constant = Double.parseDouble(str);
	  	  	q.predicates.addElement((Predicate)pred);
	  	  	q.type &= Query.THRESHOLD;
	  	  	start = queryString.indexOf("and ",start);
	  	  }
	  	
	  	return q;
	  	
	  }  // parseQuery	
	
	public void injectQuery (int queryID, String queryString)
	  {
	  	Query q = parseQuery(queryID,queryString);	  	
	  	
	  	if (q == null)
	  	  {
	  	    System.err.println("Fail to install Query " + queryID + ".");
	  	  	System.exit(1);
	  	  }	  		
	  	
	  	int i,j;
	  	for (i=0;i<m;i++)
	  	  for (j=0;j<n;j++)
	  	    eductNetwork[i][j].installQuery(q);  		  	
	  	  	
	  }  // injectQuery
	
	public void nodeTermination (Node node)
	  {
	  	int i,j,k;
	  	Query q;	  	
	  	for (i=0;i<node.queries.size();i++)
	  	  {
	  	  	q = (Query)node.queries.elementAt(i);
	  	  	PrintWriter logFileWriter = (PrintWriter)node.queryLogs.elementAt(i);
	  	  	
	  	    logFileWriter.println();
	  	  	logFileWriter.println("Average computational cost: " + (q.sumCT/q.count) + " ms.");
	  	  	logFileWriter.println("Maximum computational cost: " + q.maxCT + " ms.");
	  	  	logFileWriter.println("Minimum computational cost: " + q.minCT + " ms.");
	  	  	logFileWriter.println();
	  	  	
	  	  	logFileWriter.println();
	  	  	logFileWriter.println("Average memory requirment: " + (q.sumSR/q.count) + " bytes.");
	  	  	logFileWriter.println("Maximum memory requirment: " + q.maxSR + " bytes.");
	  	  	logFileWriter.println("Minimum memory requirment: " + q.minSR + " bytes.");
	  	  	logFileWriter.println();
	  	  	
	  	  	logFileWriter.println("Total # bytes transmitted: " + q.traffic + ".");
	  	  	logFileWriter.println();
	  	  	
	  	  	logFileWriter.flush();	  	  		  	  	
	  	  	logFileWriter.close();
	  	  	
	  	  	((Query)eductNetwork[0][0].queries.elementAt(i)).traffic += q.traffic;
	  	  }
	  	
	  	// System.out.println("Node " + node.id + " stops running.");
	  	overallLogFileWriter.println("Node " + node.id + " stops running.");
	  	overallLogFileWriter.flush();
	  	
	  	terminatedNodeNum ++;
	  	
	  	if (terminatedNodeNum == m*n)
	  	  {
	  	  	overallLogFileWriter.println();
	  	  	for (i=0;i<eductNetwork[0][0].queries.size();i++)
	  	  	  {
	  	  	  	q = (Query)eductNetwork[0][0].queries.elementAt(i);
	  	  	  		  	  	  	
	  	  	  	overallLogFileWriter.println("Total # bytes transmitted for Query " + 	q.id + ": " + q.traffic + ".");
	  	  	  	
	  	  	   	double traffic = 0;  	  	  	
	  	  	  	for (j=0;j<m;j++)
	  	  	      for (k=0;k<n;k++)
	  	  	        traffic += ((Query)eductNetwork[j][k].queries.elementAt(i)).count*ATTR_SIZE*q.attrNum*eductNetwork[j][k].depth;	  	  	        
	  	  	    overallLogFileWriter.println("Total # bytes needed to be transmitted for Query " + q.id + " if a server-side naive approach is adopted: " + traffic + ".");
	  	  	    overallLogFileWriter.println();
	  	  	    
	  	  	    double sumCT = 0, maxCT = 0, minCT = 0;
	  	  	    double sumSR = 0, maxSR = 0, minSR = 0;
	  	  	    int count = m*n-1;
	  	  	  	for (j=0;j<m;j++)
	  	  	  	  for (k=0;k<n;k++)
	  	  	  	    {
	  	  	  	      q = (Query)eductNetwork[j][k].queries.elementAt(i);
	  	  	  	      
	  	  	  	      if (q.count != 0)
	  	  	  	        {
	  	  	  	      	  sumCT += q.sumCT / q.count;
	  	  	  	      	  maxCT += q.maxCT;
	  	  	  	      	  minCT += q.minCT;
	  	  	  	      	  sumSR += q.sumSR / q.count;
	  	  	  	      	  maxSR += q.maxSR;
	  	  	  	      	  minSR += q.minSR;	  	  	  	      	    
	  	  	  	        }
	  	  	  	       else
	  	  	  	         count--;
	  	  	  	     }	  	  	  	    	  	  	  	    	
	  	  	  	sumCT /= count; maxCT /= count; minCT /= count;
	  	  	  	sumSR /= count; maxSR /= count; minSR /= count;
	  	  	  	overallLogFileWriter.println("Average computational cost for Query " + q.id + ": " + sumCT + " ms.");
	  	  	  	overallLogFileWriter.println("Maximum computational cost for Query " + q.id + ": " + maxCT + " ms.");
	  	  	  	overallLogFileWriter.println("Minimum computational cost for Query " + q.id + ": " + minCT + " ms.");
	  	  	  	overallLogFileWriter.println();
	  	  	  	
	  	  	  	overallLogFileWriter.println("Average memory requirement for Query " + q.id + ": " + sumSR + " bytes.");
	  	  	  	overallLogFileWriter.println("Maximum memory requirement for Query " + q.id + ": " + maxSR + " bytes.");
	  	  	  	overallLogFileWriter.println("Minimum memory requirement for Query " + q.id + ": " + minSR + " bytes.");
	  	  	  	
	  	  	  	overallLogFileWriter.println();
	  	  	  	overallLogFileWriter.flush();	  	  	    
	  	  	  }	  	  	
	  	  	overallLogFileWriter.flush();	  	  	
	  	  	
	  	  	System.out.println("The simulation terminates without error.");
	  	  	
	  	  	overallLogFileWriter.println("The simulation terminates without error.");
	  	  	overallLogFileWriter.println();
	  	  	overallLogFileWriter.flush();
	  	  	
	  	  	overallLogFileWriter.close();
	  	  	
	  	  	System.exit(0);
	  	  }
	  }
	
	public static void main (String argv[])
	  {
	  	try
	  	  {
	  	  	System.out.println("");
	        System.out.println("\t\t\t   Educt Simulator Version 0.11");
	        System.out.println("");	        
	        
	        Educt educt = new Educt(); 	        
	        
	        BufferedReader queryFileReader = new BufferedReader(new FileReader(queryFile));
	      
	        /** 
	         * When Educt starts, it reads a query file containing a number of query specifications (plain SQL texts)
      	       and installs them to the nodes in the network */
      	    System.out.println("Installing queries from the query file " + queryFile + " ... "); 
		    System.out.println();
		          
	        int queryID = 1;  // the IDs of the queries start from 1, 0 is reserved for system usage
	        String queryString;
	        while ( ((queryString = queryFileReader.readLine()) != null) && (queryID<=numQueries))
		      {
		        if (queryString.trim().equals("")) continue;
		      
		        System.out.println("Query " + String.valueOf(queryID) + ": " + queryString);
		        System.out.println("");
		        
		        queryString.trim();
		        queryString = queryString + " ";
		        queryString = queryString.toLowerCase(); // convert the query string to lower case
		        queryString.replace('\t',' ');
		        queryString.replace('\n',' ');
		        queryString.replace('\r',' ');
		        queryString.replaceAll("  ", " ");
		    
		        educt.injectQuery(queryID, queryString);
		       
		        queryID++;		    		    
		      } 
		  
		    System.out.println("Finish installing the queries."); 
		    System.out.println();		  
		    
		    System.out.println("The simulation starts ...");
		    
		    educt.overallLogFileWriter.println("The simulation starts ...");
		    educt.overallLogFileWriter.println();		    
		    educt.overallLogFileWriter.flush();
		    
		    while (true)
		      {
		      	int c = System.in.read();
		      	if (c=='q') // terminates the simulation manually by pressing the 'q' key
		      	  {
		      	  	System.out.println("The simulation terminates without error.");
		      	  	
		      	  	educt.overallLogFileWriter.println("The simulation terminates without error.");
		      	  	educt.overallLogFileWriter.flush();
		      	  	educt.overallLogFileWriter.close();
		      	  	
		      	  	System.exit(0);
		      	  }
		      }      
	  	  }
	  	catch (Exception e) {e.printStackTrace();}	  	
	  	
	  }  //main
	  	
}