/*

* Copyright (c) 2006, The Hong Kong University of Science and Technology (HKUST)
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without 
* modification, are permitted provided that the following conditions 
* are met:
*
*   * Redistributions of source code must retain the above copyright notice, 
*     this list of conditions,the authors and the following disclaimer.
*   * Redistributions in binary form must reproduce the above copyright notice,
*     this list of conditions,the authors and the following disclaimer in
*     the documentation and/or other materials provided with the distribution.
*   * Neither the name of the university nor the names of its 
*     contributors may be used to endorse or promote products derived from 
*     this software without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
* AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
* IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
* ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
* LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
* CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
* SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
* INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
* ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
* POSSIBILITY OF SUCH DAMAGE.
*
*/

import java.io.*;
import java.util.*;

/** 
 * A node in our simulated sensor network.
 *
 * @see Educt
 **/
public class Node {
	
	Educt educt;
	
	int id;
	
	int loc_x, loc_y;
	
	int depth;	
	
	Vector queries = new Vector (10,5);
	
	Vector sensors = new Vector (10,5);
	
	Vector caches = new Vector (10,5);
	
	Vector queryLogs = new Vector (10,5);
			
	CacheManager cm;		
	
	static final int CACHE_LIMIT = Integer.MAX_VALUE; //1024;  // unit: byte
			

	Node (Educt educt, int id)
	  {
	  	this.educt = educt;	  	
	  		  	
	  	this.id = id;	  		  	
	  	
	  	// create the sensors on this node
	  	int i;
	  	for (i=0;i<Educt.attributes.length;i++)
	  	  {
	  	  	if (Educt.attributes[i].equals("nodeid") || Educt.attributes[i].equals("loc_x") || Educt.attributes[i].equals("loc_y") || Educt.attributes[i].equals("depth"))
	  	  	  continue;
	  	  	
	  	  	Sensor s = new Sensor(this, Educt.attributes[i]);
	  	  	
	  	  	sensors.addElement((Sensor)s);
	  	  }
	  	  
	  	/**
	  	 * create the cache manager on this node
	  	 * a cache will only be created when it must be used by some query
	  	 *
	  	 */	  	
	  	cm = new CacheManager(this);	  	
	  }
	
	public void installQuery (Query q)
	  {
	  	Query qt = ((Query)q).copy();
	  	
	  	queries.addElement(qt);	  		  	
	  	
	  	qt.maxCT=Double.MIN_VALUE; qt.minCT=Double.MAX_VALUE; qt.sumCT=0;
	  	qt.maxSR=Double.MIN_VALUE; qt.minSR=Double.MAX_VALUE; qt.sumSR=0;
	  	qt.count=0; 
	  	qt.traffic=0; qt.eventNo=1;
	  	
	  	q = qt;  	
	  	
	  	try
	  	  {
	  	  	String logFileName = q.logDirName + "node" + String.valueOf(id) + ".log";
	  	  	PrintWriter logFileWriter = new PrintWriter(new FileWriter(logFileName));
	  	  	logFileWriter.println("Query id: " + q.id);
	  	  	logFileWriter.println("Query SQL text: " + q.queryText);
	  	  	if (q.type != Query.THRESHOLD)
	  	  	  logFileWriter.println("Method: type = " + q.M.type + " mode = " + q.M.mode);
	  	  	else
	  	  	  logFileWriter.println("No embedded method.");
	  	  	logFileWriter.println("Node " + id + ": loc_x = " + loc_x + " loc_y = " + loc_y + " depth = " + depth);
	  	  	logFileWriter.println();
	  	  	logFileWriter.flush();
	  	  	queryLogs.add((PrintWriter)logFileWriter);
	  	  }
	  	catch (Exception e) {e.printStackTrace();}
	  	
	  	int i;
	  	if (q.type==Query.THRESHOLD)
	  	  {
	  	  	// we assume there is only one predicate in a threshold-based query
	  	  	String attribute = ((Predicate)q.predicates.elementAt(0)).attribute;
	  	  	
	  	  	for (i=0;i<caches.size();i++)
	  	  	  if (attribute.equals(((Cache)caches.elementAt(i)).attribute))
	  	  	    break;
	  	  	if (i != caches.size())
	  	  	  return;
	  	  	
	  	  	Cache cache = new Cache(this,attribute);
	  	  	
	  	  	cache.E = 0;
	  	  	cache.IT = 0;
	  	  	cache.SP = Query.DEFAULT_SAMPLE_PERIOD;
	  	  	cache.SPAN = 1;
	  	  	
	  	  	caches.addElement((Cache)cache);  	  	
	  	  	
	  	  	cache.start();	  	
	  	  	
	  	  	return;
	  	  }	  	
	  	
	  	for (i=0;i<caches.size();i++)
	  	  if (q.M.attribute.equals(((Cache)caches.elementAt(i)).attribute))  // there is already a cache for the attribute involved
	  	    {
	  	      Cache cache = (Cache)caches.elementAt(i);
	  	      if (cache.E > q.M.epsilon)
	  	        cache.E = q.M.epsilon;	
	  	      if (cache.SP > q.M.sp)
	  	        cache.SP = q.M.sp;
	  	      if (cache.SPAN < q.M.span)
	  	        cache.SPAN = q.M.span;
	  	  	  	  	  	    
	  	  	  cache.methods.addElement((EDMethod)q.M);  	  	  
	  	    	
	  	      break;	
	  	    }
	  	    
	  	if (i == caches.size())  // no cache for the attribute involved
	  	  {
	  	  	Cache cache = new Cache(this,q.M.attribute);
	  	  	
	  	  	cache.E = q.M.epsilon;
	  	  	cache.IT = 0;
	  	  	cache.SP = q.M.sp;
	  	  	cache.SPAN = q.M.span;	  	  	
	  	  	
	  	  	caches.addElement((Cache)cache);
	  	  	
	  	  	cache.methods.addElement((EDMethod)q.M);
	  	  	
	  	  	cache.start();
	  	  }  	
	  	
	  }   // installQuery
	  
	public void executeQuery (RP s, Query q)
	  {
	  	int i, j;
	  	PrintWriter logFileWriter;	   
	  	
	  	for (i=0;i<queries.size();i++)
	  	  if (q==((Query)queries.elementAt(i)))
	  	    break;
	  	if (i==queries.size())
	  	  return;	  	
	  	logFileWriter = (PrintWriter)queryLogs.elementAt(i);
	  	
	  	//System.out.println(id);
	  	//System.exit(0);	  	
	  
	  	double attrV [] = new double [q.attrNum];
	  	String attribute, line, str;
	  	Cache cache;
	  	Sensor sensor;
	  	int mark;
	  	for (i=0;i<q.attrNum;i++)
	  	  {
	  	    attribute = q.attributes[i];
	  	      	
	  	    if (attribute.equals("nodeid"))
	  	      {attrV[i] = id; continue;}
	  	    else
	  	      if (attribute.equals("loc_x"))
	  	        {attrV[i] = loc_x; continue;}
	  	      else
	  	      	 if (attribute.equals("loc_y"))
	  	      	   {attrV[i] = loc_y; continue;}
	  	      	 else
	  	      	   if (attribute.equals("depth"))
	  	      	     {attrV[i] = depth; continue;}
	  	      	
	  	    for (j=0;j<caches.size();j++)
	  	      {
	  	      	cache = (Cache)caches.elementAt(j);
	  	      	if (attribute.equals(cache.attribute) && cache.currentTimestamp >= s.timestamp)
	  	      	  {
	  	      	  	attrV[i] = cache.currentReading; 
	  	      	  	break;
	  	      	  }
	  	      }
	  	    if (j != caches.size()) continue;
	  	      for (j=0;j<sensors.size();j++)
	  	      	 {
	  	      	    sensor = (Sensor)sensors.elementAt(j);
	  	      	  	if (attribute.equals(sensor.attribute))
	  	      	  	  {
	  	      	  	  	line = sensor.sample(s.timestamp - q.M.sp, q.M.sp);
	  	      	  	  	if (line == null)
	  	      	  	  	  {educt.nodeTermination(this); return;}
	  	                mark = line.indexOf(" ");
	  	                if (mark == -1)
	  	                  mark = line.indexOf("\t");
	  	                str = line.substring(mark);
	  	                attrV[i] = Double.parseDouble(str);
	  	      	  	  	break;
	  	      	  	  }
	  	      	  }
	  	  }   // for	  	    
	  	    
	  	/** We assume an attribute appears in the WHERE clause must also appear in the SELECT clause */
	  	for (i=0;i<q.predicates.size();i++)
	  	  {
	  	    String attr = ((Predicate)q.predicates.elementAt(i)).attribute;
	  	      	
	  	    for (j=0;j<q.attrNum;j++)
	  	      if (attr.equals(q.attributes[j]))
	  	        if (((Predicate)q.predicates.elementAt(i)).evaluate(attrV[j])==false)
	  	          return;	  	            	  	           
	  	  }
	  	    
	  	logFileWriter.println("Event detection point #" + (q.eventNo++) + ":");
	  	logFileWriter.println("Timestamp: " + s.timestamp);
	  	logFileWriter.print("Query result: ");
	  	for (j=0;j<q.attrNum;j++)
	  	  {
	  	  	logFileWriter.print(q.attributes[j] + " = ");
	  	  	logFileWriter.println(attrV[j] + " ");
	  	  }
	  	logFileWriter.flush();	
	  	  
	  	transmitQueryResult(q);
	  	  
	  }  //executeQuery			  
	  
	public void transmitQueryResult(Query q)
	  {
	  	int i;
	  	PrintWriter logFileWriter;
	  	
	  	for (i=0;i<queries.size();i++)
	  	  if (q.equals((Query)queries.elementAt(i)))
	  	    break;
	  	if (i == queries.size())
	  	  return;
	  	logFileWriter = (PrintWriter)queryLogs.elementAt(i);
	  	
	  	int byteCount = Educt.ATTR_SIZE*q.attrNum*depth;	  	
	  	logFileWriter.println("Query Result is transmitted to the base station.");
	  	logFileWriter.println("# bytes transmitted: " + byteCount + ".");
	  	logFileWriter.println();
	  	
	  	q.traffic += byteCount;  	  	
	  	 	
	  }  //transmitQueryResult  	
}